/*
 * Softcam plugin to VDR (C++)
 *
 * This code is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#ifndef ___SC_H
#define ___SC_H

#include <vdr/thread.h>

#define MAXDVBAPI 4
#define MAXSCCAPS 10

class cDevice;
class cCam;

// ----------------------------------------------------------------

struct CamInfo {
  cCam *cam;
  bool shutUp;
  };

// ----------------------------------------------------------------

class cScSetup {
public:
  int AutoUpdate;
  int ScCaps[MAXSCCAPS];
  int ConcurrentFF;
  int ConcurrentSoft;
public:
  cScSetup(void);
  void CheckConfig(void);
  };

extern cScSetup ScSetup;

// ----------------------------------------------------------------

class cSoftCAM {
friend class cCam;
private:
  static struct CamInfo camInfo[MAXDVBAPI];
  static cMutex camLock;
  static int foundKeys, newKeys;
  static int softCsaStreams;
  //
public:
  static cCam *Init(const cDevice *dev);
  static cCam *FindCAM(const cDevice *dev);
  static bool CapCheck(int n);
  static bool Load(const char *cfgdir);
  static void Save(void);
  static void ShutDown(void);
  static void HouseKeeping(void);
  static void StoreConfig(bool AsIs);
  //
  static void FoundKey(void) { foundKeys++; }
  static void NewKey(void) { newKeys++; }
  static void KeyStats(int &fk, int &nk) { fk=foundKeys; nk=newKeys; }
  static char *CurrKeyStr(int CardNum, int num);
  static bool Active(void);
  static bool ConcurrentAllowed(void);
  static void AddConcurrent(void);
  static void DelConcurrent(void);
  };

#endif // ___SC_H
