
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

static const char * const ops[] = {
//         0x00    0x01    0x02    0x03    0x04    0x05    0x06    0x07    0x08    0x09    0x0a    0x0b    0x0c    0x0d    0x0e    0x0f
/* 0x00 */ "BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR","BRSET","BRCLR",
/* 0x10 */ "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", "BSET", "BCLR", 
/* 0x20 */ "BRA",  "BRN",  "BHI",  "BLS",  "BCC",  "BCS",  "BNE",  "BEQ",  "BHCC", "BHCS", "BPL",  "BMI",  "BMC",  "BMS",  "BIL",  "BIH",
/* 0x30 */ "NEG",  "op31", "op32", "COM",  "LSR",  "op35", "ROR",  "ASR",  "ASL",  "ROL",  "DEC",  "op3b", "INC",  "TST",  "SWAP", "CLR",  
/* 0x40 */ "NEG",  "op41", "MUL",  "COM",  "LSR",  "op45", "ROR",  "ASR",  "ASL",  "ROL",  "DEC",  "op4b", "INC",  "TST",  "SWAP", "CLR",  
/* 0x50 */ "NEG",  "op51", "MUL",  "COM",  "LSR",  "op55", "ROR",  "ASR",  "ASL",  "ROL",  "DEC",  "op5b", "INC",  "TST",  "SWAP", "CLR",  
/* 0x60 */ "NEG",  "op61", "op62", "COM",  "LSR",  "op65", "ROR",  "ASR",  "ASL",  "ROL",  "DEC",  "op6b", "INC",  "TST",  "SWAP", "CLR",  
/* 0x70 */ "NEG",  "op71", "op72", "COM",  "LSR",  "op75", "ROR",  "ASR",  "ASL",  "ROL",  "DEC",  "op7b", "INC",  "TST",  "SWAP", "CLR",  
/* 0x80 */ "RTI",  "RTS",  "op82", "SWI",  "POPA", "POP%c","POPC", "op87", "PUSHA","PUSH%c","PUSHC","op8b","op8c", "op8d", "STOP", "WAIT", 
/* 0x90 */ "pre90","pre91","pre92","T%2$c%1$c","T%cS","TAS","TS%c","TA%c", "CLC",  "SEC",  "CLI",  "SEI",  "RSP",  "NOP",  "TSA",  "T%cA", 
/* 0xa0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "opa7", "EOR",  "ADC",  "ORA",  "ADD",  "opac", "BSR",  "LD%c", "opaf", 
/* 0xb0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "STA",  "EOR",  "ADC",  "ORA",  "ADD",  "JMP",  "JSR",  "LD%c", "ST%c", 
/* 0xc0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "STA",  "EOR",  "ADC",  "ORA",  "ADD",  "JMP",  "JSR",  "LD%c", "ST%c", 
/* 0xd0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "STA",  "EOR",  "ADC",  "ORA",  "ADD",  "JMP",  "JSR",  "LD%c", "ST%c", 
/* 0xe0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "STA",  "EOR",  "ADC",  "ORA",  "ADD",  "JMP",  "JSR",  "LD%c", "ST%c", 
/* 0xf0 */ "SUB",  "CMP",  "SBC",  "CP%c", "AND",  "BIT",  "LDA",  "STA",  "EOR",  "ADC",  "ORA",  "ADD",  "JMP",  "JSR",  "LD%c", "ST%c", 
  };

static unsigned char mem[64*1024];
static unsigned short pc;
static bool indirect;
static unsigned char x, y;

#define Get(x)        (mem[(x)])
#define HILO(ea)      ((Get(ea)<<8)+Get(ea+1))
#define branch(x)     Branch()
#define DEBUG_EMU
#define de(x)         { (x); }

void Branch(void)
{
  unsigned char off=Get(pc);
  if(indirect) {
    de(printf("[%02x] -> ",off))
    off=Get(off);
    }
  unsigned short npc=pc+off+1;
  if(off&0x80) npc-=0x100; // gcc fixup. take care of sign
  de(printf("%04x ",npc))
  pc++;
}

int Run(unsigned short stopPc)
{
  de(printf("-pc- -bytes-  -inst-\n"))
  while (1) {
    if(pc>=stopPc) return 0;

    de(printf("%04x ",pc))
    unsigned short spc=pc;

    unsigned char *ex=&x;
    indirect=false;
    unsigned char ins=Get(pc++);
    char xs='X';
    int ilen=1;
 
    // check pre-byte (ST7 extention)
    switch(ins) {
      case 0x91: // use Y register with indirect addr mode
        indirect=true;
        // fall through
      case 0x90: // use Y register
        ex=&y; de(xs='Y')
        ins=Get(pc++); ilen++;
        break;
      case 0x92: // use indirect addr mode
        indirect=true;
        ins=Get(pc++); ilen++;
        break;
      }

    switch(ins>>4) {
      case 0x8:
      case 0x9:
        break;
      case 0x1:                 // bit
      case 0x2:			// branch
      case 0xA:			// immediate
      case 0x3:			// short
      case 0xB:
      case 0x6:			// short indexed
      case 0xE:
        ilen++; break;
      case 0xC:			// long
      case 0xD:			// long indexed
        if(!indirect) ilen+=2;
        else ilen++;
        break;
      case 0x0:			// bit
        ilen+=2; break;
      }
    for(int i=0; i<4; i++) {
      if(i<ilen) printf("%02x",Get(spc+i));
      else printf("  ");
      }

    {
    char str[8];
    snprintf(str,sizeof(str),ops[ins],xs,xs^1);
    printf(" %-5s ",str);
    }

    // address decoding
    unsigned short ea=0;
    switch(ins>>4) {
      case 0x2:			// no or special address mode
      case 0x8:
      case 0x9:
        break;
      case 0xA:			// immediate
        de(printf("#%02x ",Get(pc)))
        ea=pc++; break;
      case 0x3:			// short
      case 0xB:
        ea=Get(pc++);
        if(!indirect) {		// short direct
          }
        else {			// short indirect
          de(printf("[%02x] -> ",ea))
          ea=Get(ea);
          }
        de(printf("%02x ",ea))
        break;
      case 0xC:			// long
        if(!indirect) {		// long direct
          ea=HILO(pc); pc+=2;
          }
        else {			// long indirect
          ea=Get(pc++);
          de(printf("[%02x] -> ",ea))
          ea=HILO(ea);
          }
        de(printf("%04x ",ea))
        break;
      case 0xD:			// long indexed
        if(!indirect) {		// long direct indexed
          ea=HILO(pc); pc+=2;
          de(printf("(%04x",ea))
          }
        else {			// long indirect indexed
          ea=Get(pc++);
          de(printf("([%02x]",ea))
          ea=HILO(ea);
          de(printf(",%c) -> (%04x",xs,ea))
          }
        ea+=*ex;
        de(printf(",%c) -> %04x ",xs,ea))
        break;
      case 0x6:			// short indexed
      case 0xE:
        ea=Get(pc++);
        if(!indirect) {		// short direct indexed
          de(printf("(%02x",ea))
          }
        else {			// short indirect indexed
          de(printf("([%02x]",ea))
          ea=Get(ea);
          de(printf(",%c) -> (%02x",xs,ea))
          }
        ea+=*ex;
        de(printf(",%c) -> %04x ",xs,ea))
        break;
      case 0x7:			// indexed
      case 0xF:
        ea=*ex;
        de(printf("(%c) -> %02x ",xs,ea))
        break;
      case 0x4:			// inherent A
        de(printf("A "))
        break;
      case 0x5:			// inherent X/Y
        de(printf("%c ",xs))
        break;
      case 0x0:			// bit
      case 0x1:
        ea=Get(pc++);
        if(!indirect) {
          de(printf("%02x",ea))
          }
        else {
          de(printf("[%02x]",ea))
          ea=Get(ea);
          indirect=false; // don't use indirect mode in case this is a bit branch
          }
        break;
      }

    // command decoding
    switch(ins) {
      case 0x00 ... 0x0F: // BRSET BRCLR
        {
        int bit=(ins&0x0F)>>1;
        de(printf(",#%x,",bit))
        branch((ins&0x01) ? !cc.c:cc.c);
        break;
        }
      case 0x10 ... 0x1F: // BSET BCLR
        {
        int bit=(ins&0x0F)>>1;
        de(printf(",#%x",bit))
        break;
        }
      case 0x20: // BRA
        branch(true); break;
      case 0x21: // BRN
        branch(false); break;
      case 0x22: // BHI
        branch(!cc.c && !cc.z); break;
      case 0x23: // BLS
        branch( cc.c ||  cc.z); break;
      case 0x24: // BCC BHS
        branch(!cc.c); break;
      case 0x25: // BCS BLO
        branch( cc.c); break;
      case 0x26: // BNE
        branch(!cc.z); break;
      case 0x27: // BEQ
        branch( cc.z); break;
      case 0x28: // BHCC
        branch(!cc.h); break;
      case 0x29: // BHCS
        branch( cc.h); break;
      case 0x2A: // BPL
        branch(!cc.n); break;
      case 0x2B: // BMI
        branch( cc.n); break;
      case 0x2C: // BMC
        branch(!cc.i); break;
      case 0x2D: // BMS
        branch( cc.i); break;
      }

    de(printf("\n"))
    }
}

int main(int argc, char *argv[])
{
  if(argc<5) {
    printf("usage: %s <addr> <bin/dump> <file> <start-pc> <end-pc>\n",argv[0]);
    return 1;
    }

  int sa=strtol(argv[1],0,0);
  if(sa<0 || sa>=0xFFFF) {
    printf("bad start address 0x%x\n",sa);
    return 1;
    }

  bool binMode;
  if(!strcasecmp(argv[2],"bin")) binMode=true;
  else if(!strcasecmp(argv[2],"dump")) binMode=false;
  else {
    printf("bad file type. Only 'bin' or 'dump' allowed\n");
    return 1;
    }

  int len=0;
  FILE *f=fopen(argv[3],"r");
  if(f) {
    if(binMode) {
      len=fread(&mem[sa],1,0x10000-sa,f);
      }
    else {
      char line[512];
      while(len<0x10000-sa && fgets(line,sizeof(line),f)) {
        if(!strncmp(line,"dump:",5)) continue;
        char *p=index(line,':');
        if(p) p++; else p=line;
        while(1) {
          char *np;
          int val=strtol(p,&np,16);
          if(p!=np) {
            mem[sa+len]=val; len++;
            p=np;
            }
          else break;
          }
        }
      }
    }
  else {
    printf("failed to open file %s: %s\n",argv[3],strerror(errno));
    return 1;
    }

  pc=strtol(argv[4],0,0);
  if(pc<sa || pc>=sa+len) {
    printf("bad start-pc address 0x%x\n",pc);
    return 1;
    }

  int endPc=strtol(argv[5],0,0);
  if(endPc<=pc || endPc>sa+len) {
    printf("bad end-pc address 0x%x\n",endPc);
    return 1;
    }

  Run(endPc);
  return 0;
}
